@@ -1,4 +1,4 @@
-/* $Id: tcpprep.c,v 1.38 2004/05/08 21:30:26 aturner Exp $ */
+/* $Id: tcpprep.c 878 2004-11-07 03:31:04Z aturner $ */
 
 /*
  * Copyright (c) 2001-2004 Aaron Turner.
@@ -52,10 +52,7 @@
 #include <string.h>
 #include <unistd.h>
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif /* HAVE_CONFIG_H */
-
 #include "tcpreplay.h"
 #include "cache.h"
 #include "cidr.h"
@@ -67,6 +64,8 @@
 #include "rbtree.h"
 #include "utils.h"
 #include "services.h"
+#include "sll.h"
+#include "fakepcap.h"
 
 /*
  * global variables
@@ -161,9 +160,11 @@ static void
 print_comment(char *file)
 {
     char *cachedata = NULL;
+    u_int64_t count = 0;
 
-    read_cache(&cachedata, file);
+    count = read_cache(&cachedata, file);
     printf("tcpprep args: %s\n", options.tcpprep_comment);
+    printf("Cache contains data for %llu packets\n", count);
 
     exit(0);
 }
@@ -246,16 +247,57 @@ process_raw_packets(pcap_t * pcap)
 {
     ip_hdr_t *ip_hdr = NULL;
     eth_hdr_t *eth_hdr = NULL;
+    struct sll_header *sll_hdr = NULL;
+    struct cisco_hdlc_header *hdlc_hdr = NULL;
+    int l2len = 0;
+    u_int16_t protocol = 0;
     struct pcap_pkthdr pkthdr;
     const u_char *pktdata = NULL;
     unsigned long packetnum = 0;
+    int linktype = 0;
 #ifdef FORCE_ALIGN
     u_char ipbuff[MAXPACKET];
 #endif
 
     while ((pktdata = pcap_next(pcap, &pkthdr)) != NULL) {
         packetnum++;
-        eth_hdr = (eth_hdr_t *) pktdata;
+        eth_hdr = NULL;
+        sll_hdr = NULL;
+        ip_hdr = NULL;
+        hdlc_hdr = NULL;
+
+        linktype = pcap_datalink(pcap);
+        dbg(1, "Linktype is %s (0x%x)", 
+                pcap_datalink_val_to_description(linktype), linktype);
+        switch (linktype) {
+        case DLT_EN10MB:
+            eth_hdr = (eth_hdr_t *) pktdata;
+            l2len = LIBNET_ETH_H;
+            protocol = eth_hdr->ether_type;
+            break;
+
+        case DLT_LINUX_SLL:
+            sll_hdr = (struct sll_header *) pktdata;
+            l2len = SLL_HDR_LEN;
+            protocol = sll_hdr->sll_protocol;
+            break;
+
+        case DLT_RAW:
+            protocol = ETHERTYPE_IP;
+            l2len = 0;
+            break;
+
+        case DLT_CHDLC:
+            hdlc_hdr = (struct cisco_hdlc_header *)pktdata;
+            protocol = hdlc_hdr->protocol;
+            l2len = CISCO_HDLC_LEN;
+            break;
+
+        default:
+            errx(1, "WTF?  How'd we get here with an invalid DLT type: %s (0x%x)",
+                 pcap_datalink_val_to_description(linktype), linktype);
+            break;
+        }
 
         dbg(1, "Packet %d", packetnum);
 
@@ -273,8 +315,8 @@ process_raw_packets(pcap_t * pcap)
             }
         }
 
-        if (ntohs(eth_hdr->ether_type) != ETHERTYPE_IP) {
-            dbg(2, "Packet isn't IP: 0x%.2x", eth_hdr->ether_type);
+        if (htons(protocol) != ETHERTYPE_IP) {
+            dbg(2, "Packet isn't IP: %#0.4x", protocol);
 
             if (mode != AUTO_MODE)  /* we don't want to cache
                                      * these packets twice */
@@ -287,17 +329,17 @@ process_raw_packets(pcap_t * pcap)
          * copy layer 3 and up to our temp packet buffer
          * for now on, we have to edit the packetbuff because
          * just before we send the packet, we copy the packetbuff 
-         * back onto the pkt.data + LIBNET_ETH_H buffer
+         * back onto the pkt.data + l2len buffer
          * we do all this work to prevent byte alignment issues
          */
         ip_hdr = (ip_hdr_t *) & ipbuff;
-        memcpy(ip_hdr, (pktdata + LIBNET_ETH_H), (pkthdr.caplen - LIBNET_ETH_H));
+        memcpy(ip_hdr, (pktdata + l2len), (pkthdr.caplen - l2len));
 #else
         /*
          * on non-strict byte align systems, don't need to memcpy(), 
-         * just point to 14 bytes into the existing buffer
+         * just point to l2len bytes into the existing buffer
          */
-        ip_hdr = (ip_hdr_t *) (pktdata + LIBNET_ETH_H);
+        ip_hdr = (ip_hdr_t *) (pktdata + l2len);
 #endif
 
         /* look for include or exclude CIDR match */
@@ -353,7 +395,7 @@ process_raw_packets(pcap_t * pcap)
              * process ports based on their destination port
              */
             add_cache(&cachedata, 1, 
-                      check_dst_port(ip_hdr, (pkthdr.caplen - LIBNET_ETH_H)));
+                      check_dst_port(ip_hdr, (pkthdr.caplen - l2len)));
             break;
         }
 
@@ -517,8 +559,12 @@ main(int argc, char *argv[])
             parse_services(optarg);
             break;
         case 'x':
+            if (include_exclude_mode != 0)
+                errx(1, "Error: Can only specify -x OR -X");
+
             include_exclude_mode = 'x';
-            if ((xX = parse_xX_str(include_exclude_mode, optarg)) == NULL)
+            if ((include_exclude_mode = 
+                 parse_xX_str(include_exclude_mode, optarg, &xX)) == 0)
                 errx(1, "Unable to parse -x: %s", optarg);
             if (include_exclude_mode & xXPacket) {
                 xX_list = (LIST *) xX;
@@ -528,8 +574,12 @@ main(int argc, char *argv[])
             }
             break;
         case 'X':
+            if (include_exclude_mode != 0)
+                errx(1, "Error: Can only specify -x OR -X");
+
             include_exclude_mode = 'X';
-            if ((xX = parse_xX_str(include_exclude_mode, optarg)) == NULL)
+            if ((include_exclude_mode = 
+                 parse_xX_str(include_exclude_mode, optarg, &xX)) == 0)
                 errx(1, "Unable to parse -X: %s", optarg);
             if (include_exclude_mode & xXPacket) {
                 xX_list = (LIST *) xX;
@@ -596,6 +646,14 @@ main(int argc, char *argv[])
         errx(1, "Error opening file: %s", errbuf);
     }
 
+    if ((pcap_datalink(pcap) != DLT_EN10MB) &&
+        (pcap_datalink(pcap) != DLT_LINUX_SLL) &&
+        (pcap_datalink(pcap) != DLT_RAW) &&
+        (pcap_datalink(pcap) != DLT_CHDLC)) {
+        errx(1, "Unsupported pcap DLT type: 0x%x", pcap_datalink(pcap));
+    }
+
+
     /* do we apply a bpf filter? */
     if (options.bpf_filter != NULL) {
         if (pcap_compile(pcap, &bpf, options.bpf_filter,
