@@ -1,4 +1,4 @@
-/* $Id: tcpreplay.c,v 1.91 2004/05/15 21:11:50 aturner Exp $ */
+/* $Id: tcpreplay.c 884 2004-11-08 17:04:47Z aturner $ */
 
 /*
  * Copyright (c) 2001-2004 Aaron Turner, Matt Bing.
@@ -56,6 +56,7 @@
 #include "tcpdump.h"
 #include "cache.h"
 #include "cidr.h"
+#include "portmap.h"
 #include "list.h"
 #include "err.h"
 #include "do_packets.h"
@@ -64,11 +65,13 @@
 #include "replay_live.h"
 #include "utils.h"
 #include "edit_packet.h"
+#include "fakepcap.h"
 
 struct options options;
 char *cachedata = NULL;
 CIDR *cidrdata = NULL;
 CIDRMAP *cidrmap_data1 = NULL, *cidrmap_data2 = NULL;
+PORTMAP *portmap_data = NULL;
 struct timeval begin, end;
 u_int64_t bytes_sent, failed, pkts_sent;
 char *cache_file = NULL, *intf = NULL, *intf2 = NULL;
@@ -119,7 +122,7 @@ main(int argc, char *argv[])
 
     while ((ch =
             getopt(argc, argv,
-                   "bc:C:De:f:Fhi:I:j:J:k:K:l:L:m:MnN:o:Op:Pr:Rs:S:t:Tu:Vw:W:x:X:12:"
+                   "bc:C:De:f:Fhi:I:j:J:k:K:l:L:m:MnN:o:Op:Pr:Rs:S:t:Tu:Vw:W:x:X:12:4:"
 #ifdef HAVE_TCPDUMP
                    "vA:"
 #endif
@@ -166,7 +169,7 @@ main(int argc, char *argv[])
             break;
         case 'I':              /* primary dest mac */
             mac2hex(optarg, options.intf1_mac, sizeof(options.intf1_mac));
-            if (memcmp(options.intf1_mac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", optarg);
             break;
         case 'j':              /* secondary interface */
@@ -174,17 +177,17 @@ main(int argc, char *argv[])
             break;
         case 'J':              /* secondary dest mac */
             mac2hex(optarg, options.intf2_mac, sizeof(options.intf2_mac));
-            if (memcmp(options.intf2_mac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf2_mac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", optarg);
             break;
         case 'k':              /* primary source mac */
             mac2hex(optarg, options.intf1_smac, sizeof(options.intf1_smac));
-            if (memcmp(options.intf1_smac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf1_smac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", optarg);
             break;
        case 'K':              /* secondary source mac */
             mac2hex(optarg, options.intf2_smac, sizeof(options.intf2_smac));
-            if (memcmp(options.intf2_smac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf2_smac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", optarg);
             break;
         case 'l':              /* loop count */
@@ -272,11 +275,11 @@ main(int argc, char *argv[])
             break;
         case 'S':              /* enable live replay mode w/ snaplen */
             options.sniff_snaplen = atoi(optarg);
-            if ((options.sniff_snaplen < 0) || (options.sniff_snaplen > 65535)) {
+            if ((options.sniff_snaplen < 0) || (options.sniff_snaplen > MAX_SNAPLEN)) {
                 errx(1, "Invalid snaplen: %d", options.sniff_snaplen);
             }
             else if (options.sniff_snaplen == 0) {
-                options.sniff_snaplen = 65535;
+                options.sniff_snaplen = MAX_SNAPLEN;
             }
             break;
         case 't':              /* MTU */
@@ -361,12 +364,13 @@ main(int argc, char *argv[])
                 errx(1, "Error: Can only specify -x OR -X");
 
             include_exclude_mode = 'x';
-            if ((xX = parse_xX_str(include_exclude_mode, optarg)) == NULL)
+            if ((include_exclude_mode = 
+                 parse_xX_str(include_exclude_mode, optarg, &xX)) == 0)
                 errx(1, "Unable to parse -x: %s", optarg);
+
             if (include_exclude_mode & xXPacket) {
                 xX_list = (LIST *) xX;
-            }
-            else if (!include_exclude_mode & xXBPF) {
+            } else if (! (include_exclude_mode & xXBPF)) {
                 xX_cidr = (CIDR *) xX;
             }
             break;
@@ -375,14 +379,15 @@ main(int argc, char *argv[])
                 errx(1, "Error: Can only specify -x OR -X");
 
             include_exclude_mode = 'X';
-            if ((xX = parse_xX_str(include_exclude_mode, optarg)) == NULL)
+            if ((include_exclude_mode = 
+                 parse_xX_str(include_exclude_mode, optarg, &xX)) == 0)
                 errx(1, "Unable to parse -X: %s", optarg);
+
             if (include_exclude_mode & xXPacket) {
                 xX_list = (LIST *) xX;
-            }
-            else if (!include_exclude_mode & xXBPF) {
+            } else {
                 xX_cidr = (CIDR *) xX;
-            }
+            } 
             break;
         case '1':              /* replay one packet at a time */
             options.one_at_a_time = 1;
@@ -395,6 +400,12 @@ main(int argc, char *argv[])
             l2enabled = 1;
             l2len = read_hexstring(optarg, l2data, L2DATALEN);
             break;
+        case '4':
+            options.rewriteports = 1;
+
+            if (! parse_portmap(&portmap_data, optarg))
+                errx(1, "Invalid port mapping");
+            break;
         default:
             usage();
         }
@@ -493,8 +504,8 @@ main(int argc, char *argv[])
         /*
          * we also can't rewrite macs for non-802.3
          */
-        if ((memcmp(options.intf1_mac, NULL_MAC, 6) == 0) ||
-            (memcmp(options.intf2_mac, NULL_MAC, 6) == 0))
+        if ((memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0) ||
+            (memcmp(options.intf2_mac, NULL_MAC, LIBNET_ETH_H) == 0))
             errx(1,
                  "You can't rewrite destination MAC's with non-802.3 frames");
 
@@ -671,7 +682,8 @@ replay_live(char *iface, int l2enabled, char *l2data, int l2len)
                          options.bpf_optimize, 0) != 0) {
             errx(1, "Error compiling BPF filter: %s", pcap_geterr(pcap));
         }
-        pcap_setfilter(pcap, &bpf);
+        if (pcap_setfilter(pcap, &bpf) != 0)
+            errx(1, "Unable to apply BPF filter: %s", pcap_geterr(pcap));
     }
 
     do_packets(NULL, pcap, linktype, l2enabled, l2data, l2len);
@@ -698,7 +710,7 @@ replay_file(char *path, int l2enabled, char *l2data, int l2len)
 #endif
 
     if ((pcapnav = pcapnav_open_offline(path)) == NULL) {
-        errx(1, "Error opening file: %s", strerror(errno));
+        errx(1, "Error opening file %s: %s", path, strerror(errno));
     }
 
     pcap = pcapnav_pcap(pcapnav);
@@ -741,31 +753,80 @@ void
 validate_l2(char *name, int l2enabled, char *l2data, int l2len, int linktype)
 {
 
-    if (linktype != DLT_EN10MB) {
-        if (linktype == DLT_LINUX_SLL) {
-            /* if SLL, then either -2 or -I are ok */
-            if ((memcmp(options.intf1_mac, NULL_MAC, 6) == 0) && (!l2enabled)) {
-                warnx
-                    ("Unable to process Linux Cooked Socket pcap without -2 or -I: %s",
-                     name);
-                return;
-            }
 
-            /* if using dual interfaces, make sure -2 or -J is set */
-            if (options.intf2 &&
-                ((!l2enabled) ||
-                 (memcmp(options.intf2_mac, NULL_MAC, 6) == 0))) {
-                warnx
-                    ("Unable to process Linux Cooked Socket pcap with -j without -2 or -J: %s",
-                     name);
+    dbg(1, "Linktype is %s\n", pcap_datalink_val_to_description(linktype));
+
+    switch (linktype) {
+    case DLT_EN10MB:
+        /* nothing to do here */
+        break;
+
+    case DLT_LINUX_SLL:
+        
+        /* single output mode */
+        if (! options.intf2) {
+            /* if SLL, then either -2 or -I are ok */
+            if ((memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0) && (!l2enabled)) {
+                warnx("Unable to process pcap without -2 or -I: %s", name);
                 return;
             }
         }
-        else if (!l2enabled) {
-            warnx("Unable to process non-802.3 pcap without layer 2 data: %s",
-                  name);
+        
+        /* dual output mode */
+        else {
+            /* if using dual interfaces, make sure -2 or -J & -I) is set */
+            if (((memcmp(options.intf2_mac, NULL_MAC, LIBNET_ETH_H) == 0) ||
+                 (memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0)) &&
+                (! l2enabled)) {
+                errx(1, "Unable to process pcap with -j without -2 or -I & -J: %s",  name);
+                return;
+            }
+        }            
+        break;
+            
+    case DLT_CHDLC:
+        /* Cisco HDLC (used at least for SONET) */
+        /* 
+         * HDLC has a 4byte header, a 2 byte address type (0x0f00 is unicast
+         * is all I know) and a 2 byte protocol type
+         */
+            
+        /* single output mode */
+        if (! options.intf2) {
+            /* Need either a full l2 header or -I & -k */
+            if (((memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0) || 
+                 (memcmp(options.intf1_smac, NULL_MAC, LIBNET_ETH_H) == 0)) &&
+                (! l2enabled)) {
+                errx(1, "Unable to process pcap without -2 or -I and -k: %s", name);
+                return;
+            }
+        }
+        
+        /* dual output mode */
+        else {
+            /* Need to have a l2 header or -J, -K, -I, -k */
+            if (((memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0) ||
+                 (memcmp(options.intf1_smac, NULL_MAC, LIBNET_ETH_H) == 0) ||
+                 (memcmp(options.intf2_mac, NULL_MAC, LIBNET_ETH_H) == 0) ||
+                 (memcmp(options.intf2_smac, NULL_MAC, LIBNET_ETH_H) == 0)) &&
+                (! l2enabled)) {
+                errx(1, "Unable to process pcap with -j without -2 or -J, -I, -K & -k: %s", name);
+                return;
+            }
+        }
+        break;
+        
+    case DLT_RAW:
+        if (!l2enabled) {
+            errx(1, "Unable to process pcap without -2: %s",  name);
             return;
         }
+        break;
+
+    default:
+        errx(1, "validate_l2(): Unsupported datalink type: %s (0x%x)", 
+             pcap_datalink_val_to_description(linktype), linktype);
+        break;
     }
 
     /* calculate the maxpacket based on the l2len, linktype and mtu */
@@ -774,7 +835,7 @@ validate_l2(char *name, int l2enabled, char *l2data, int l2len, int linktype)
         dbg(1, "Using custom L2 header to calculate max frame size");
         maxpacket = options.mtu + l2len;
     }
-    else if ((linktype == DLT_EN10MB) || (linktype == DLT_LINUX_SLL)) {
+    else if (linktype == DLT_EN10MB) {
         /* ethernet */
         dbg(1, "Using Ethernet to calculate max frame size");
         maxpacket = options.mtu + LIBNET_ETH_H;
@@ -850,12 +911,12 @@ configfile(char *file)
         }
         else if (ARGS("primary_mac", 2)) {
             mac2hex(argv[1], options.intf1_mac, sizeof(options.intf1_mac));
-            if (memcmp(options.intf1_mac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf1_mac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", argv[1]);
         }
         else if (ARGS("primary_smac", 2)) {
             mac2hex(argv[1], options.intf1_smac, sizeof(options.intf1_smac));
-            if (memcmp(options.intf1_smac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf1_smac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", argv[1]);
         }
         else if (ARGS("second_intf", 2)) {
@@ -863,12 +924,12 @@ configfile(char *file)
         }
         else if (ARGS("second_mac", 2)) {
             mac2hex(argv[1], options.intf2_mac, sizeof(options.intf2_mac));
-            if (memcmp(options.intf2_mac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf2_mac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", argv[1]);
         }
         else if (ARGS("second_smac", 2)) {
             mac2hex(argv[1], options.intf2_smac, sizeof(options.intf2_smac));
-            if (memcmp(options.intf2_smac, NULL_MAC, 6) == 0)
+            if (memcmp(options.intf2_smac, NULL_MAC, LIBNET_ETH_H) == 0)
                 errx(1, "Invalid mac address: %s", argv[1]);
         }
         else if (ARGS("loop", 2)) {
@@ -901,6 +962,12 @@ configfile(char *file)
                     errx(1, "Invalid secondary NAT string");
             }
         }
+        else if (ARGS("portmap", 2)) {
+            options.rewriteports = 1;
+
+            if (! parse_portmap(&portmap_data, argv[1]))
+                errx(1, "Invalid port mapping");
+        }
 #ifdef HAVE_PCAPNAV
         else if (ARGS("offset", 2)) {
             options.offset = strtoull(argv[1], NULL, 0);
@@ -958,11 +1025,11 @@ configfile(char *file)
         }
         else if (ARGS("sniff_snaplen", 2)) {
             options.sniff_snaplen = atoi(argv[1]);
-            if ((options.sniff_snaplen < 0) || (options.sniff_snaplen > 65535)) {
+            if ((options.sniff_snaplen < 0) || (options.sniff_snaplen > MAX_SNAPLEN)) {
                 errx(1, "Invalid sniff snaplen: %d", options.sniff_snaplen);
             }
             else if (options.sniff_snaplen == 0) {
-                options.sniff_snaplen = 65535;
+                options.sniff_snaplen = MAX_SNAPLEN;
             }
         }
         else if (ARGS("packetrate", 2)) {
@@ -979,7 +1046,8 @@ configfile(char *file)
                 errx(1,
                      "Error: Can only specify include (-x) OR exclude (-X) ");
             include_exclude_mode = 'x';
-            if ((xX = parse_xX_str(include_exclude_mode, argv[1])) == NULL)
+            if ((include_exclude_mode = 
+                 parse_xX_str(include_exclude_mode, argv[1], &xX)) == 0)
                 errx(1, "Unable to parse include: %s", optarg);
             if (include_exclude_mode & xXPacket) {
                 xX_list = (LIST *) xX;
@@ -993,7 +1061,8 @@ configfile(char *file)
                 errx(1, "Error: Can only specify include (-x) OR exclude (-X)");
 
             include_exclude_mode = 'X';
-            if ((xX = parse_xX_str(include_exclude_mode, argv[1])) == NULL)
+            if ((include_exclude_mode =
+                 parse_xX_str(include_exclude_mode, argv[1], &xX)) == 0)
                 errx(1, "Unable to parse exclude: %s", optarg);
             if (include_exclude_mode & xXPacket) {
                 xX_list = (LIST *) xX;
@@ -1133,6 +1202,7 @@ usage(void)
            "-X <match>\t\tSend all the packets except those specified\n"
            "-1\t\t\tSend one packet per key press\n"
            "-2 <datafile>\t\tLayer 2 data\n"
+           "-4 <PORT1:PORT2,...>\tRewrite port numbers\n"
            "<file1> <file2> ...\tFile list to replay\n");
     exit(1);
 }
@@ -1181,7 +1251,7 @@ init(void)
     cache_bit = cache_byte = 0;
 
     if (fcntl(STDERR_FILENO, F_SETFL, O_NONBLOCK) < 0)
-        errx(1, "Unable to set STDERR to non-blocking");
+        warnx("Unable to set STDERR to non-blocking: %s", strerror(errno));
 }
 
 void
